import os
import math
import sys
import csv
import time
import argparse
import pathlib
import configparser
import script_tools
import script_unpack 

parser = argparse.ArgumentParser(description='scans a folder of doozy-j scripts to see how they all load each other (using 0x67)')
parser.add_argument('-f', '--filter', type=str, help='only check scripts with filenames containing the given string')
parser.add_argument('-i', '--config', type=pathlib.Path, help='config file for script_tools.py')
parser.add_argument('input_path', type=pathlib.Path, help='path to original script folder')
parser.add_argument('output_path', type=pathlib.Path, help='path to output script edge list')

args = parser.parse_args()

def scan_file(args, graphEdges, filenameAsHexInt):
    start_time = time.time()
    input_path = args.input_path

    config_path = args.config
    config = configparser.ConfigParser()
    if config_path and config_path.is_file() and config_path.exists(): 
        config.read(config_path)

    if input_path.is_file() and input_path.exists(): 
        inputFilenameOnly = input_path.name
        print("Beginning scanning of %s..." % (inputFilenameOnly))
        in_file = open(input_path, "rb")
        file_bytes = in_file.read()
        # zeroes are not truncated because sometimes trailing 00s are used in arguments

        readingTableOfContents = True
        tableOfContentsCount = config.getint(inputFilenameOnly.upper(), 'toc_section_count', fallback=-1)
        prevTableOfContentsIndex = 0
        prevOpcodes = []
        byte_index = 0
        while byte_index < len(file_bytes):
            b = file_bytes[byte_index]
            skipByteCheck = False

            if readingTableOfContents:
                indexHex = bytearray(file_bytes[(byte_index + 0) : (byte_index + 2)])
                indexInt = int.from_bytes(indexHex, 'little')

                # TOC is inferred by being an increasing 2-byte word OR if the first index is a reasonable number
                if tableOfContentsCount == -1:
                    if indexInt < 255 and prevTableOfContentsIndex == 0: # idk a reasonable inference
                        tableOfContentsCount = indexInt/2

                    elif indexInt > prevTableOfContentsIndex:
                        prevTableOfContentsIndex = indexInt

                        skipByteCheck = True
                        byte_index += 2
                    else: 
                        readingTableOfContents = False
                
                if tableOfContentsCount > 0:
                    prevTableOfContentsIndex = indexInt
                    tableOfContentsCount -= 1

                    skipByteCheck = True
                    byte_index += 2
                elif tableOfContentsCount == 0:
                    readingTableOfContents = False

            if skipByteCheck:
                continue

            prevOpcodes.append(b)

            if b == 0x67 or b == 0x65:
                stringEndHex = bytearray(file_bytes[(byte_index + 1) : (byte_index + 3)])
                stringEndInt = int.from_bytes(stringEndHex, 'little')
                edge = ["%04x" % filenameAsHexInt, "%04x" % stringEndInt]
                graphEdges.append(edge)

            opcodeInfo = script_tools.informationForOpcodeAtLocation(file_bytes, byte_index, config, inputFilenameOnly, None, {})
            if opcodeInfo.valid:
                byte_index += opcodeInfo.totalLength()
            else:
                if b >= 0x0: 
                    print("WARNING: unexpected opcode %02x at position %d. Skipping." % (b, byte_index))

                    prevOpcodes = prevOpcodes[-15:]
                    opcodeHex = []
                    for opcode in prevOpcodes:
                        opcodeHex.append("%02x" % opcode)
                    print("         Previous opcodes: %s" % opcodeHex)
                byte_index += 1

    print("Scanning of %s complete in %s seconds.\n" % (inputFilenameOnly, time.time() - start_time))

def scan_all(args):
    start_time = time.time()

    graphEdges = []
    for filename in os.listdir(args.input_path):
        if len(filename) < 2:
            continue # lazy skip for "."

        if args.filter and args.filter not in filename:
            continue

        filename2 = filename.split("_")[1] # files are in format ####_000000FF; we only want the hex
        filenameWithoutExt = os.path.splitext(filename2)[0]
        filenameAsHexInt = int(filenameWithoutExt, 16)

        f = pathlib.Path(os.path.join(args.input_path, filename))
        f2 = pathlib.Path(os.path.join(args.output_path, filename) + ".csv")
        unpackArgs = argparse.Namespace(input_path=f, output_path=f2)
        if args.config:
            unpackArgs.config = args.config
        
        scan_file(unpackArgs, graphEdges, filenameAsHexInt)

        # save after every file just in case
        output_file = open(args.output_path, 'w', encoding='utf-8')
        csv_writer = csv.writer(output_file, delimiter='|', quotechar='{', quoting=csv.QUOTE_MINIMAL, lineterminator='\n')
        for row in graphEdges:
            csv_writer.writerow(row)
        output_file.close()

    print("All script scanning complete in %s seconds." % (time.time() - start_time))

scan_all(args)