import os
import math
import sys
import csv
import time
import argparse
import pathlib
import configparser
import difflib

import script_tools

def main():
    parser = argparse.ArgumentParser(description='modifies a Zero4 Champ Doozy-J (PSX) script using a CSV translation file')
    parser.add_argument('-c', '--debug-clean', action='store_true', help='ignores output_path, effectively performing a repack with zero changes. for debugging')
    parser.add_argument('-d', '--debug-diff', action='store_true', help='does not actually output script. instead diffs the bytes of output and output_path to check for changes')
    parser.add_argument('-i', '--config', type=pathlib.Path, help='config file containing overrides and other info')
    parser.add_argument('input_csv', type=pathlib.Path, help='path to translation CSV (format: start pos, end pos, J string, E string)')
    parser.add_argument('input_path', type=pathlib.Path, help='path to original script')
    parser.add_argument('output_path', type=pathlib.Path, help='path to output the final script to (using the same filename as original_script)')
    args = parser.parse_args()
    repack(args)

def repack(args): 
    start_time = time.time()

    input_csv = args.input_csv
    input_path = args.input_path
    output_path = args.output_path

    config = configparser.ConfigParser()
    try:
        if args.config:
            config_path = args.config
        if config_path and config_path.is_file() and config_path.exists(): 
            config.read(config_path)
    except AttributeError:
        pass

    # thanks https://stackoverflow.com/questions/2422177/python-how-can-i-replace-full-width-characters-with-half-width-characters
    HALF2FULL = dict((i, i + 0xFEE0) for i in range(0x21, 0x7F))
    HALF2FULL[0x20] = 0x3000

    class StringReplaceStruct:
        startPos = None
        sizeDifference = None
        newStringBytes = None

    # TODO: figure out what to do with scripts that are near the char limit. using single byte might reallllly help

    if input_path.is_file() and input_path.exists(): 
        inputFilenameOnly = input_path.name
        orig_file = open(input_path, "rb")
        file_bytes = bytearray(orig_file.read())
        # zeroes are not truncated because sometimes trailing 00s are used in arguments

        input_file = open(input_csv, 'r', encoding='utf-8')
        csv_reader = csv.reader(input_file, delimiter='|', quotechar='{', quoting=csv.QUOTE_MINIMAL, lineterminator='\n')

        stringReplacements = []

        # find all addresses in the code
        addressLocations = []        

        readingTableOfContents = True
        tableOfContentsCount = config.getint(inputFilenameOnly.upper(), 'toc_section_count', fallback=-1)
        prevTableOfContentsIndex = 0
        prevOpcodes = []
        byte_index = 0
        while byte_index < len(file_bytes):
            b = file_bytes[byte_index]
            skipByteCheck = False

            if readingTableOfContents:
                indexHex = bytearray(file_bytes[(byte_index + 0) : (byte_index + 2)])
                indexInt = int.from_bytes(indexHex, 'little')

                # TOC is inferred by being an increasing 2-byte word OR if the first index is a reasonable number
                if tableOfContentsCount == -1:
                    if indexInt < 255 and prevTableOfContentsIndex == 0: # idk a reasonable inference
                        tableOfContentsCount = indexInt/2

                    elif indexInt > prevTableOfContentsIndex:
                        prevTableOfContentsIndex = indexInt
                        addressLocations.append(byte_index)

                        skipByteCheck = True
                        byte_index += 2
                    else: 
                        readingTableOfContents = False
                
                if tableOfContentsCount > 0:
                    prevTableOfContentsIndex = indexInt
                    addressLocations.append(byte_index)
                    tableOfContentsCount -= 1

                    skipByteCheck = True
                    byte_index += 2
                elif tableOfContentsCount == 0:
                    readingTableOfContents = False

            if skipByteCheck:
                continue

            prevOpcodes.append(b)
            opcodeInfo = script_tools.informationForOpcodeAtLocation(file_bytes, byte_index, config, inputFilenameOnly, None, {})
            if opcodeInfo.valid:
                for val in opcodeInfo.location_indexes:
                    addressLocations.append(byte_index + val)
                byte_index += opcodeInfo.totalLength()
            else:
                if b >= 0x0: 
                    print("WARNING: unexpected opcode %02x at position %d. Skipping." % (b, byte_index))

                    prevOpcodes = prevOpcodes[-15:]
                    opcodeHex = []
                    for opcode in prevOpcodes:
                        opcodeHex.append("%02x" % opcode)
                    print("         Previous opcodes: %s" % opcodeHex)
                byte_index += 1

        lastStringEndPos = -1
        lastStringWasInsertion = False
        for row in csv_reader:
            if args.debug_clean: # ignore all rows if so
                continue

            if len(row) < 5:
                print("WARNING: Invalid CSV row %s" % row)
                continue


            finalString = row[4]
            # process the final string
            finalString.strip()
            charaReplacementList = [
                ['"', '¨'], # TODO: find a better quote mark
                ["'", '′'], # TODO: lol
                ["-", 'ー'],
            ]

            if len(finalString) > 0: # if there is a string to replace
                # make sure the charas in brackets [ ] don't get fullwidth'd, and also remove the brackets
                finalStringIngame = ""
                bracketOpen = False
                for char in finalString:
                    if char == '[':
                        bracketOpen = True
                    elif char == ']':
                        bracketOpen = False
                    else:
                        if bracketOpen == False:
                            for charaTuple in charaReplacementList:
                                if char == charaTuple[0]:
                                    char = charaTuple[1]
                            finalStringIngame += char.translate(HALF2FULL) # inline fullLen for time
                        else:
                            finalStringIngame += char
                finalStringBytes = finalStringIngame.encode(encoding='shift-jis')

                if row[0] == "line" or row[0] == "dialog": 
                    # this is a new line that wasn't in the original
                    if row[0] == "line":
                        additionalBytes = bytearray.fromhex('95')
                    elif row[0] == "dialog":
                        additionalBytes = bytearray.fromhex('97 03 00 00 00 93 01 00 00 00 92 95')

                    finalStringBytes = bytearray(finalStringBytes)

                    byteName = "95"
                    startPos = lastStringEndPos # TODO: check for invalid (< 0) pos)
                    endPos = startPos + len(finalStringBytes) + len(additionalBytes) + 2

                    # calculate the proper end address for this and inject it
                    offsetNeededForAddress = script_tools.findOffsetForPositionWithRewrites(startPos, stringReplacements, {})
                    realEndAddress = endPos + offsetNeededForAddress
                    if lastStringWasInsertion == True: # hack to account for the previous insertion, which we will extend
                        replacement = stringReplacements[-1] 
                        realEndAddress += replacement.sizeDifference
                    addressBytes = bytearray(realEndAddress.to_bytes(2, 'little')) 

                    finalStringBytes = additionalBytes + addressBytes + finalStringBytes

                    # TODO: build structure where you can check if the last insertion place matches and extend it
                    if lastStringWasInsertion == True:
                        # extend the previous insertion
                        replacement = stringReplacements[-1]
                        replacement.sizeDifference += len(finalStringBytes)
                        replacement.newStringBytes += finalStringBytes
                    else:
                        replacement = StringReplaceStruct()
                        replacement.startPos = startPos
                        replacement.sizeDifference = len(finalStringBytes)
                        replacement.newStringBytes = finalStringBytes
                        stringReplacements.append(replacement)

                    # lastStringEndPos = endPos
                    lastStringWasInsertion = True
                else: # normal line
                    startPos = int(row[0])
                    endPos = int(row[1])
                    byteName = row[2]
                    origString = row[3]

                    origStringLength = endPos-startPos
                    finalStringSizeDifference = len(finalStringBytes) - origStringLength

                    replacement = StringReplaceStruct()
                    replacement.startPos = startPos
                    replacement.sizeDifference = finalStringSizeDifference
                    replacement.newStringBytes = finalStringBytes
                    stringReplacements.append(replacement)

                    lastStringEndPos = endPos
                    lastStringWasInsertion = False
            else:
                # hack to ensure you can put new "line" things after untranslated strings. though you shouldn't
                endPos = int(row[1])
                lastStringEndPos = endPos
                lastStringWasInsertion = False


        # write the string rewrites
        # the string pointers will be shifted later as shown below
        totalDifference = 0 # running total of how much the offset has shifted
        for replacement in stringReplacements:
            sizeDifference = replacement.sizeDifference
            finalStringBytes = replacement.newStringBytes

            startPos = replacement.startPos + totalDifference
            
            # resize array to fit
            if sizeDifference < 0:
                file_bytes[startPos:startPos+abs(sizeDifference)] = [] 
            elif sizeDifference > 0:
                for i in range(0, sizeDifference):
                    file_bytes.insert(startPos, 0)
            # TODO: all of this is probably bad for perf

            for strIndex in range(0, len(finalStringBytes)):
                file_bytes[startPos+strIndex] = finalStringBytes[strIndex]
            totalDifference += sizeDifference

        # perform the jump rewrites
        memoization = {}
        for addressLocation in addressLocations:
            offsetNeededForAddress = script_tools.findOffsetForPositionWithRewrites(addressLocation, stringReplacements, memoization)
            realAddressLocation = addressLocation + offsetNeededForAddress
            script_tools.rewriteLocationAtPositionWithNewOffsetIfNeeded(realAddressLocation, file_bytes, stringReplacements, memoization)

        input_file.close()

        # check and print the difference in byte size
        orig_file = open(str(output_path) + "\\" + inputFilenameOnly, "rb")
        original_bytes = bytearray(orig_file.read())
        orig_file.close()

        # trim excess zeroes
        file_bytes = file_bytes.rstrip(b'\0')
        orig_file.close()
        original_bytes = original_bytes.rstrip(b'\0')

        if len(original_bytes) != len(file_bytes):
            print("Filesize change from %d to %d bytes." % (len(original_bytes), len(file_bytes)))
        if script_tools.getBlockSizeOfData(original_bytes) != script_tools.getBlockSizeOfData(file_bytes):
            print("Blocksize change from %d to %d blocks." % (script_tools.getBlockSizeOfData(original_bytes), script_tools.getBlockSizeOfData(file_bytes)))

        # check and print the difference in byte value
        # TODO: this doesn't do smart diffing. sorry. need to use difflib for that
        bytediffs = []
        new_byte_index = 0
        max_index = max(len(file_bytes), len(original_bytes))
        while new_byte_index < max_index:
            if new_byte_index >= len(file_bytes):
                bytediffs.append([new_byte_index, b'\0', original_bytes[new_byte_index]])
            elif new_byte_index >= len(original_bytes):
                bytediffs.append([new_byte_index, file_bytes[new_byte_index], b'\0'])
            elif file_bytes[new_byte_index] != original_bytes[new_byte_index]:
                bytediffs.append([new_byte_index, file_bytes[new_byte_index], original_bytes[new_byte_index]])
            new_byte_index += 1
        if len(bytediffs) > 0:
            print("%d bytes changed." % len(bytediffs))
            diff_index = 0
            while diff_index < 20 and diff_index < len(bytediffs):
                print("Diff at byte %d: change from %02x to %02x" % (bytediffs[diff_index][0], bytediffs[diff_index][2], bytediffs[diff_index][1]))
                diff_index += 1

        # then actually write the change if appropriate
        if args.debug_diff == False:
            file_bytes = script_tools.padBytesToBlockSize(file_bytes)

            output_file = open(str(output_path) + "\\" + inputFilenameOnly, "wb")
            output_file.write(file_bytes)
            output_file.close()

    print("Script repacking complete in %s seconds." % (time.time() - start_time))

if __name__ == '__main__':
    main()
