import os
import math
import sys
import csv
import time
import argparse
import pathlib
import configparser
import keyboard

# thanks https://stackoverflow.com/questions/2422177/python-how-can-i-replace-full-width-characters-with-half-width-characters

BLOCK_SIZE = 2048 
SCAN_AHEAD_COUNT = 9999

def fullLen(s):
    '''
    Convert all ASCII characters to the full-width counterpart.
    '''
    return str(s).translate(HALF2FULL)  

def padBytesToBlockSize(data):
    while len(data) % BLOCK_SIZE != 0:
        data += b"\0"
    return data

def getBlockSizeOfData(data):
    blocks = int(math.ceil(1.0 * len(data) / BLOCK_SIZE))
    return blocks

def findOffsetForPositionWithRewrites(pos, rewrites, memo):
    if memo == None or rewrites == None:
        return 0

    offset = 0
    if pos in memo:
        return memo[pos]

    for rewrite in rewrites:
        if rewrite.startPos < pos:
            offset += rewrite.sizeDifference

    memo[pos] = offset
    return offset

def rewriteLocationAtPositionWithNewOffsetIfNeeded(pos, file_bytes, rewrites, memo):
    jumpLocHex = bytearray(file_bytes[pos : pos + 2])
    jumpLocInt = int.from_bytes(jumpLocHex, 'little')

    offsetNeeded = findOffsetForPositionWithRewrites(jumpLocInt, rewrites, memo)
    if offsetNeeded != 0:
        jumpLocInt += offsetNeeded
        newLocBytes = bytearray(jumpLocInt.to_bytes(2, 'little')) 
        for strIndex in range(0, len(newLocBytes)):
            file_bytes[pos+strIndex] = newLocBytes[strIndex]

class OpcodeInformation:
    valid = False
    arguments = []
    location_indexes = [] # relative to start of opcode
    reachedEndOfFile = False

    startByteIndex = 0
    description = ""

    def __init__(self):
        self.valid = False
        self.arguments = []
        self.location_indexes = []

    def addArgument(self, byte_size):
        self.valid = True
        self.arguments.append(byte_size)

    def getArgumentByteIndex(self, argument):
        offset = 1
        i = 0
        while i < argument:
            offset += self.arguments[i]
            i += 1
        return self.startByteIndex + offset

    def addLocation(self):
        self.valid = True
        self.location_indexes.append(self.totalLength())
        self.addArgument(2)

    def markNoOp(self):
        self.valid = True

    def totalLength(self):
        return 1+sum(self.arguments)

    def markEndOfFile(self):
        self.reachedEndOfFile = True

def getHexStringFor4ByteValueAtOpCodeArgument(file_bytes, opcodeInfo, argumentIndex):
    # TODO: safety
    pos = opcodeInfo.getArgumentByteIndex(argumentIndex)
    hexBytes = bytearray(file_bytes[pos : pos + 4])
    intValue = int.from_bytes(hexBytes, 'little')
    return ("%04x" % intValue)

def rewriteAddressInOpcode(opcodeInfo, file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization):
    rewriteLocationAtPositionWithNewOffsetIfNeeded(byte_index + opcodeInfo.totalLength(), file_bytes, stringReplacements, memoization)
    opcodeInfo.addLocation()

# manual prompt to input the length of a switch
def switchPrompt(file_bytes, byte_index, configSection, configKey, config, memoization):
    thinksAreValidCases = [] # TODO: convert to set
    lowestValid = 9999
    i = 0
    while i < 255:
        # might get burned by using configSection == inputFilenameOnly
        couldBeValidOpcode = scanAheadForSafety(SCAN_AHEAD_COUNT, file_bytes, byte_index + i * 2, config, configSection, memoization, 0)
        if (couldBeValidOpcode):
            thinksAreValidCases.append(i)
            if i < lowestValid:
                lowestValid = i
        i += 1

    selectionMade = False
    selectionValue = -1
    selectionOffset = 0
    while selectionMade != True:
        print("Switch at 0x%04x:" % byte_index)
        print("Press number key to select first non-switch address, [m] to pick the minimum valid (**) switch, +/- to seek, and Esc to skip")

        i = 0
        while i < 9:
            pos = byte_index + (selectionOffset + i) * 2 
            hexBytes = bytearray(file_bytes[pos : pos + 2])
            addressValue = int.from_bytes(hexBytes, 'big') # this is intentionally backwards to preserve the byte order

            byteString = "%04x" % addressValue
            bonusString = ""
            if (selectionOffset+i) in thinksAreValidCases:
                bonusString = "**"
            if (selectionOffset == 0):
                print("[%d] %s %s" % ((i+1), byteString, bonusString))
            else:
                print("[%d (%d)] %s %s" % ((i+1), (selectionOffset+i), byteString, bonusString))

            i += 1

        keyPress = None
        while keyPress == None or keyPress.event_type != "up":
            keyPress = keyboard.read_event(False)
        keyCode = keyPress.name
        if keyCode == "esc":
            selectionMade = True
            selectionValue = -1
        # could do this better but I'm lazy
        if keyCode == "1" or keyCode == "2" or keyCode == "3" or keyCode == "4" or keyCode == "5" or keyCode == "6" or keyCode == "7" or keyCode == "8" or keyCode == "9":
            selectionMade = True
            selectionValue = int(keyCode)-1+selectionOffset
        if keyCode == "m" and lowestValid < 9999:
            selectionMade = True
            selectionValue = lowestValid
        if keyCode == "-":
            selectionOffset -= 9
            if selectionOffset < 0:
                selectionOffset = 0
        if keyCode == "=":
            selectionOffset += 9
            # TODO: safety for reaching eof
    
    if selectionValue >= 0:
        if config.has_section(configSection) == False:
            config.add_section(configSection)

        config.set(configSection, configKey, str(selectionValue))

# scan that many opcodes ahead and confirm that they are valid
def scanAheadForSafety(scanCount, file_bytes, byte_index, config, inputFilenameOnly, memoization, depth):
    debugScan = False

    opcodeKey = "scanAheadKey_" + str(byte_index)

    if byte_index >= len(file_bytes): # dunno what's up but that's invalid
        return False
    if depth > 100: # if you've gone this far I don't like it and I don't know why TODO
        return False

    if opcodeKey in memoization:
        return memoization[opcodeKey]

    if debugScan:
        print("Try %d (depth %d) *******************" % (byte_index, depth))
    while scanCount > 0:
        if debugScan:
            print("Scan %d (%d left)" % (byte_index, scanCount))
        opcodeInfo = informationForOpcodeAtLocation(file_bytes, byte_index, config, inputFilenameOnly, None, memoization, depth+1, quiet=True)
        if opcodeInfo.reachedEndOfFile:
            memoization[opcodeKey] = True
            return True
        elif opcodeInfo.valid:
            byte_index += opcodeInfo.totalLength()
        else:
            memoization[opcodeKey] = False
            return False
        scanCount -= 1
    memoization[opcodeKey] = True
    return True

def informationForOpcodeAtLocation(file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization, depth=0, quiet=False):
    b = file_bytes[byte_index]
    opcodeInfo = OpcodeInformation()
    opcodeInfo.startByteIndex = byte_index

    if b == 0x00: # could be the trailing zeroes
        zeroMemoKey = "zeroKey_" + str(byte_index)

        theIndex = byte_index
        isAllZero = True

        if zeroMemoKey in memoization:
            isAllZero = memoization[zeroMemoKey]
        else:
            while theIndex < len(file_bytes):
                if file_bytes[theIndex] != 0x00:
                    isAllZero = False
                    break
                theIndex += 1
            #memoization[zeroMemoKey] = isAllZero # TODO: don't use this memoization for now
        if isAllZero:
            opcodeInfo.markEndOfFile()
            opcodeInfo.markNoOp() # so errors don't happen

    if b == 0x64: # Jump
        rewriteAddressInOpcode(opcodeInfo, file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization)

    elif b == 0x65: # Jump to file (clears stack?)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

        # TODO: consider annotating opcodes and branches
        #if quiet == False:
        #    fileString = getHexStringFor4ByteValueAtOpCodeArgument(file_bytes, opcodeInfo, 0)
        #    fileIndex = getHexStringFor4ByteValueAtOpCodeArgument(file_bytes, opcodeInfo, 1)
        #    opcodeInfo.description = "jump to file (clear stack) with value %s and parameter %s" % (fileString, fileIndex)
        #    print(opcodeInfo.description)
        #    print(byte_index)

    elif b == 0x66: # Jump (pushes return address on stack?)
        rewriteAddressInOpcode(opcodeInfo, file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization)

    elif b == 0x67: # Jump to file
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0x68: # Return
        opcodeInfo.markNoOp()

    elif b == 0x69: # Jump if 8007a0a8 lowest bit is 1
        rewriteAddressInOpcode(opcodeInfo, file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization)

    elif b == 0x6A: # Jump if 8007a0a8 2nd lowest bit is 1
        rewriteAddressInOpcode(opcodeInfo, file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization)

    elif b == 0x6B: # Jump if 8007a0a8 lowest bit is 0
        rewriteAddressInOpcode(opcodeInfo, file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization)

    elif b == 0x6C: # Jump if 8007a0a8 2nd lowest bit is 0
        rewriteAddressInOpcode(opcodeInfo, file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization)

    elif b == 0x6D: # Jump if equal
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        rewriteAddressInOpcode(opcodeInfo, file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization)

    elif b == 0x6E: # Jump if not equal
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        rewriteAddressInOpcode(opcodeInfo, file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization)

    elif b == 0x6F: # Jump if A >= B
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        rewriteAddressInOpcode(opcodeInfo, file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization)

    elif b == 0x70: # Jump if A > B
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        rewriteAddressInOpcode(opcodeInfo, file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization)

    elif b == 0x71: # Jump if A <= B
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        rewriteAddressInOpcode(opcodeInfo, file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization)

    elif b == 0x72: # Jump if A < B
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        rewriteAddressInOpcode(opcodeInfo, file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization)

    elif b == 0x73:
        opcodeInfo.addArgument(4)

    elif b == 0x74:
        opcodeInfo.addArgument(4)

    elif b == 0x75:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0x76:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0x77:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0x78:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0x79:
        opcodeInfo.addArgument(4)

    elif b == 0x7A: # Jump if ??? nonzero
        opcodeInfo.addArgument(4)
        rewriteAddressInOpcode(opcodeInfo, file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization)

    elif b == 0x7B: # Jump if ??? zero
        opcodeInfo.addArgument(4)
        rewriteAddressInOpcode(opcodeInfo, file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization)

    elif b == 0x7C: # Change expression
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0x7D:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0x7E:
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4)

    elif b == 0x7F:
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4)

    elif b == 0x80:
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4)

    elif b == 0x81:
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4)

    elif b == 0x82:
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4)

    elif b == 0x83:
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4)

    elif b == 0x84:
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4)

    elif b == 0x85:
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4)

    elif b == 0x86:
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4)

    elif b == 0x87:
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4)

    elif b == 0x88:
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4)

    elif b == 0x89:
        opcodeInfo.addArgument(4)

    elif b == 0x8A:
        opcodeInfo.addArgument(4)

    elif b == 0x8B: # ? Doesn't do much visibly
        opcodeInfo.markNoOp()

    elif b == 0x8C:
        opcodeInfo.markNoOp()

    elif b == 0x8D:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0x8E:
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4)

    elif b == 0x8F:
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4)

    elif b == 0x90:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0x91:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0x92: # Clear draw box
        opcodeInfo.markNoOp()

    elif b == 0x93: # Set draw box
        opcodeInfo.addArgument(4)

    elif b == 0x94:
        opcodeInfo.markNoOp()

    elif b == 0x95: # Draw string to current box
        stringEndHex = bytearray(file_bytes[(byte_index + 1) : (byte_index + 3)])
        stringEndInt = int.from_bytes(stringEndHex, 'little')
        rewriteAddressInOpcode(opcodeInfo, file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization)
        stringEndHex = bytearray(file_bytes[(byte_index + 1) : (byte_index + 3)])
        stringEndInt = int.from_bytes(stringEndHex, 'little')

        if stringEndInt < byte_index:
            if quiet == False:
                print("WARNING: Negative-length string detected at %s. Skipping" % (byte_index))
            opcodeInfo.valid = False
        elif stringEndInt > byte_index + 0xFF * 3:
            if quiet == False:
                print("WARNING: Really big string detected at %s. Skipping" % (byte_index))
            opcodeInfo.valid = False
        else:
            opcodeInfo.addArgument(stringEndInt-byte_index-3) # jump over the entire string

    elif b == 0x96: # Used to prompt for player choice
        startIndex = byte_index

        switchOverride = config.getint(inputFilenameOnly.upper(), ('%02x' % byte_index).upper() + '_switch_count', fallback=-1)

        if switchOverride == -1 and config.has_option("general", "switch_prompt") and config.getboolean("general", "switch_prompt") and quiet == False:
            switchPrompt(file_bytes, byte_index+33, inputFilenameOnly.upper(), ('%02x' % byte_index).upper() + '_switch_count', config, memoization)
            switchOverride = config.getint(inputFilenameOnly.upper(), ('%02x' % byte_index).upper() + '_switch_count', fallback=-1)

        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

        # basically if this cannot be a valid opcode, assume it is an address
        # Assume there is at least one case because switch statements need at least 1
        cases = 1
        rewriteAddressInOpcode(opcodeInfo, file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization)

        if switchOverride > 0:
            while cases < switchOverride:
                rewriteAddressInOpcode(opcodeInfo, file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization)
                cases += 1  
        else:      
            while cases < 255 and scanAheadForSafety(SCAN_AHEAD_COUNT, file_bytes, byte_index + opcodeInfo.totalLength(), config, inputFilenameOnly, memoization, depth) == False:
                rewriteAddressInOpcode(opcodeInfo, file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization)
                cases += 1
            if quiet == False:
                print("0x96 switch at %d (0x%02x) with unspecified case count, fuzzily detected %d" % (byte_index, byte_index, cases))
            if cases > 255:
                if quiet == False:
                    print("0x96 switch at %d (0x%02x) has more than 255 cases, that's weird" % (byte_index, byte_index))
                opcodeInfo.valid = False

    elif b == 0x97: # Used to wait for player input
        opcodeInfo.addArgument(4)

    elif b == 0x98: # very weird
        if quiet == False:
            print("WARNING: unknown behavior opcode 98 at position %d. Skipping." % (byte_index))
        opcodeInfo.valid = False

    elif b == 0x99:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0x9A: # "reverbOff"
        opcodeInfo.markNoOp()

    elif b == 0x9B:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0x9C:
        opcodeInfo.addArgument(4)

    elif b == 0x9D:
        # read 3 params forward, then keep skipping 9 bytes until you hit FF
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        byte_index += 13
        while byte_index < len(file_bytes) and file_bytes[byte_index] != 0xFF:
            opcodeInfo.addArgument(9)
            byte_index += 9
        opcodeInfo.addArgument(1)

        if byte_index >= len(file_bytes):
            if quiet == False:
                print("WARNING: 9D opcode at %d trying to read past end of data." % (byte_index))
            opcodeInfo.valid = False

    elif b == 0x9E:
        opcodeInfo.addArgument(4)

    elif b == 0x9F:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0xA0:
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4)

    elif b == 0xA1: # TODO: 99% sure this is a string lol
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        rewriteAddressInOpcode(opcodeInfo, file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization)

        # seems to be a location where Some Data is read until, so skip to after that location
        stringEndHex = bytearray(file_bytes[(byte_index + 5) : (byte_index + 7)])
        stringEndInt = int.from_bytes(stringEndHex, 'little')

        if stringEndInt < byte_index:
            if quiet == False:
                print("WARNING: Negative-length A1 sequence detected at %s. Skipping" % (byte_index))
            opcodeInfo.valid = False
        elif stringEndInt > byte_index + 0xFF * 3:
            if quiet == False:
                print("WARNING: Really big A1 sequence detected at %s. Skipping" % (byte_index))
            opcodeInfo.valid = False
        else:
            opcodeInfo.addArgument(stringEndInt-byte_index-7)

    elif b == 0xA2:
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4)

    elif b == 0xA3:
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4)

    elif b == 0xA4:
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4)

    elif b == 0xA5:
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0xA6: # possibly init stuff?
        opcodeInfo.markNoOp()

    elif b == 0xA7:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0xA8: # "reverbOn"
        opcodeInfo.markNoOp()

    elif b == 0xA9:
        opcodeInfo.addArgument(4)

    elif b == 0xAA:
        opcodeInfo.addArgument(4)

    elif b == 0xAB: # unused (causes error)
        opcodeInfo.markNoOp()

    elif b == 0xAC:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0xAD:
        opcodeInfo.addArgument(4)

    elif b == 0xAE:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0xAF:
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4)

    elif b == 0xB0:
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4)

    elif b == 0xB1:
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4)

    elif b == 0xB2: # no-op
        opcodeInfo.markNoOp()

    elif b == 0xB3: # Switch block (no return)
        startOfSwitch = byte_index
        valueToSwitchHex = bytearray(file_bytes[(byte_index + 1) : (byte_index + 5)])
        valueToSwitchInt = int.from_bytes(valueToSwitchHex, 'little')
        opcodeInfo.addArgument(4)

        # we don't actually know how big the switch block is here, so gotta fuzzily guess it
        previousSwitchInt = -1
        keepChecking = True
        expectedGapThreshold = 255*2 # assume things can't be further than this far apart
        cases = 0

        switchOverride = config.getint(inputFilenameOnly.upper(), ('%02x' % byte_index).upper() + '_switch_count', fallback=-1)

        if switchOverride == -1 and config.has_option("general", "switch_prompt") and config.getboolean("general", "switch_prompt") and quiet == False:
            switchPrompt(file_bytes, byte_index+5, inputFilenameOnly.upper(), ('%02x' % byte_index).upper() + '_switch_count', config, memoization)
            switchOverride = config.getint(inputFilenameOnly.upper(), ('%02x' % byte_index).upper() + '_switch_count', fallback=-1)

        if switchOverride != -1:
            keepChecking = False

        while switchOverride > 0:
            rewriteAddressInOpcode(opcodeInfo, file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization)
            switchOverride -= 1

        while keepChecking and cases < 255 and byte_index + opcodeInfo.totalLength() < len(file_bytes):
            nextSwitchHex = bytearray(file_bytes[(byte_index + opcodeInfo.totalLength()) : (byte_index + opcodeInfo.totalLength() + 2)])
            nextSwitchInt = int.from_bytes(nextSwitchHex, 'little')

            # TODO: should not use scanAhead for writing because it will mess up when hitting 95 strings. can you just write them to config?
            couldBeValidOpcode = scanAheadForSafety(SCAN_AHEAD_COUNT, file_bytes, byte_index + opcodeInfo.totalLength(), config, inputFilenameOnly, memoization, depth)
            destinationIsValidAddress = False
            if nextSwitchInt < len(file_bytes):
                destinationIsValidAddress = scanAheadForSafety(SCAN_AHEAD_COUNT, file_bytes, nextSwitchInt, config, inputFilenameOnly, memoization, depth)
            # unused stuff:  and nextSwitchInt >= previousSwitchInt and (abs(nextSwitchInt-previousSwitchInt) < expectedGapThreshold)
            if previousSwitchInt < 0 or couldBeValidOpcode == False or destinationIsValidAddress:
                previousSwitchInt = nextSwitchInt
                rewriteAddressInOpcode(opcodeInfo, file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization)
                cases += 1
            else:
                keepChecking = False

        if switchOverride == -1:
            if quiet == False:
                print("0xB3 switch at %d (0x%02x) with unspecified case count, fuzzily detected %d" % (byte_index, byte_index, cases))
        if cases > 255:
            if quiet == False:
                print("0xB3 switch at %d (0x%02x) has more than 255 cases, that's weird" % (byte_index, byte_index))
            opcodeInfo.valid = False

    elif b == 0xB4: # Switch block
        valueToSwitchHex = bytearray(file_bytes[(byte_index + 1) : (byte_index + 5)])
        valueToSwitchInt = int.from_bytes(valueToSwitchHex, 'little')
        switchCountHex = bytearray(file_bytes[(byte_index + 5) : (byte_index + 9)])
        switchCountInt = int.from_bytes(switchCountHex, 'little')

        if switchCountInt > 0xFF:
            if quiet == False:
                print("0xB4 switch at %d (0x%02x) with too many switches, skipping" % (byte_index, byte_index))
        else:
            opcodeInfo.addArgument(4)
            opcodeInfo.addArgument(4)

            for i in range(0, switchCountInt):
                rewriteAddressInOpcode(opcodeInfo, file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization)

    elif b == 0xB5:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0xB6:
        opcodeInfo.addArgument(4)

    elif b == 0xB7:
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0xB8:
        opcodeInfo.markNoOp()

    elif b == 0xB9:
        opcodeInfo.markNoOp()

    elif b == 0xBA:
        opcodeInfo.markNoOp()

    elif b == 0xBB:
        opcodeInfo.markNoOp()

    elif b == 0xBC: # possibly init stuff?
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        if quiet != False:
            print("0xBC load at %d (0x%02x)" % (byte_index, byte_index))

    elif b == 0xBD:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0xBE:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0xBF:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0xC0:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0xC1: # Jump if ???
        opcodeInfo.addArgument(4)
        rewriteAddressInOpcode(opcodeInfo, file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization)

    elif b == 0xC2:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0xC3: # this one's weird
        if quiet == False:
            print("WARNING: unknown behavior opcode C3 at position %d. Skipping." % (byte_index))
        opcodeInfo.valid = False

    elif b == 0xC4: # unused (COPY, causes error)
        opcodeInfo.markNoOp()

    elif b == 0xC5:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0xC6:
        opcodeInfo.markNoOp()

    elif b == 0xC7:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0xC8:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0xC9: # no-op
        opcodeInfo.markNoOp()

    elif b == 0xCA: # no-op
        opcodeInfo.markNoOp()

    elif b == 0xCB: # no-op
        opcodeInfo.markNoOp()

    # TODO: everything below here is manually guessed. check function comeBackLater to confirm
    # TODO: please update sheet with this

    elif b == 0xD8:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        rewriteAddressInOpcode(opcodeInfo, file_bytes, byte_index, config, inputFilenameOnly, stringReplacements, memoization)

    elif b == 0xD9:
        opcodeInfo.addArgument(4)  # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4) 

    elif b == 0xDA:
        opcodeInfo.addArgument(4)  # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4) 

    elif b == 0xDB:
        opcodeInfo.addArgument(4)  # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4) 

    elif b == 0xDC:
        opcodeInfo.addArgument(4)  # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4) 

    elif b == 0xDD:
        opcodeInfo.addArgument(4)  # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4) 

    elif b == 0xDE:
        opcodeInfo.addArgument(4)  # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4) 

    elif b == 0xDF:
        opcodeInfo.addArgument(4)  # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4) 

    elif b == 0xE0:
        opcodeInfo.addArgument(4)  # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4) 

    elif b == 0xE1:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4) 

    elif b == 0xE2:
        opcodeInfo.markNoOp()

    elif b == 0xE9:
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4) 
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4) 

    elif b == 0xEA:
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4) 
        opcodeInfo.addArgument(4) # 4-byte type 1 (instead of type 0)
        opcodeInfo.addArgument(4) 

    elif b == 0xEB:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0xEC:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0xED:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0xEE:
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)

    elif b == 0xEF:
        opcodeInfo.addArgument(4)

    elif b == 0xF0:
        opcodeInfo.markNoOp()
        
    elif b == 0xF1:
        opcodeInfo.markNoOp()
        
    elif b == 0xF2:
        opcodeInfo.markNoOp()
        
    elif b == 0xF3:
        opcodeInfo.markNoOp()
        
    elif b == 0xF4:
        opcodeInfo.markNoOp()
        
    elif b == 0xF5: # appears to load file
        opcodeInfo.addArgument(4)
        opcodeInfo.addArgument(4)
        if quiet == False:
            print("0xF5 load at %d (0x%02x)" % (byte_index, byte_index))
        
    elif b == 0xF6: # TODO: this is weird, possibly log it for checking
        if file_bytes[byte_index + 1] == 0x00:
            opcodeInfo.addArgument(1) # the 0x00
            opcodeInfo.addArgument(4)
        else:
            opcodeInfo.addArgument(1) # a byte that is not 0x00

    return opcodeInfo
