import os
import math
import sys
import csv
import time
import argparse
import pathlib
import configparser
import script_tools

def main():
    parser = argparse.ArgumentParser(description='extracts text and logic from a Zero4 Champ Doozy-J (PSX) script into a CSV translation file')
    parser.add_argument('-i', '--config', type=pathlib.Path, help='config file containing overrides and other info')
    parser.add_argument('input_path', type=pathlib.Path, help='path to original script')
    parser.add_argument('output_path', type=pathlib.Path, help='path for the output CSV (TODO: how to imply this is a full filename)')
    args = parser.parse_args()
    unpack(args)

ignoreEmptyLines = True

def unpack(args):
    start_time = time.time()

    input_path = args.input_path
    output_path = args.output_path

    config = configparser.ConfigParser()
    try:
        if args.config:
            config_path = args.config
        if config_path and config_path.is_file() and config_path.exists(): 
            config.read(config_path)
    except AttributeError:
        pass

    if input_path.is_file() and input_path.exists(): 
        inputFilenameOnly = input_path.name
        print("Beginning unpacking of %s..." % (inputFilenameOnly))
        in_file = open(input_path, "rb")
        file_bytes = in_file.read()
        # zeroes are not truncated because sometimes trailing 00s are used in arguments

        output_file = open(output_path, 'w', encoding='utf-8')
        csv_writer = csv.writer(output_file, delimiter='|', quotechar='{', quoting=csv.QUOTE_MINIMAL, lineterminator='\n')

        readingTableOfContents = True
        tableOfContentsCount = config.getint(inputFilenameOnly.upper(), 'toc_section_count', fallback=-1)
        prevTableOfContentsIndex = 0
        prevOpcodes = []
        byte_index = 0
        while byte_index < len(file_bytes):
            b = file_bytes[byte_index]
            skipByteCheck = False

            if readingTableOfContents:
                indexHex = bytearray(file_bytes[(byte_index + 0) : (byte_index + 2)])
                indexInt = int.from_bytes(indexHex, 'little')

                # TOC is inferred by being an increasing 2-byte word OR if the first index is a reasonable number
                if tableOfContentsCount == -1:
                    if indexInt < 255 and prevTableOfContentsIndex == 0: # idk a reasonable inference
                        tableOfContentsCount = indexInt/2

                    elif indexInt > prevTableOfContentsIndex:
                        prevTableOfContentsIndex = indexInt

                        skipByteCheck = True
                        byte_index += 2
                    else: 
                        readingTableOfContents = False
                
                if tableOfContentsCount > 0:
                    prevTableOfContentsIndex = indexInt
                    tableOfContentsCount -= 1

                    skipByteCheck = True
                    byte_index += 2
                elif tableOfContentsCount == 0:
                    readingTableOfContents = False

            if skipByteCheck:
                continue

            prevOpcodes.append(b)

            if b == 0x95 or b == 0xA1: # string
                if b == 0xA1:
                    byte_index += 4 # AFAIK a 0xA1 opcode is just 0x95 with a 4-byte argument

                stringStartInt = (byte_index + 3)
                stringStartHex = stringStartInt.to_bytes(2, 'little')

                stringEndHex = bytearray(file_bytes[(byte_index + 1) : (byte_index + 3)])
                stringEndHex.reverse()
                stringEndInt = int.from_bytes(stringEndHex, 'big')

                if stringEndInt < byte_index:
                    print("WARNING: Negative-length string detected at %s. Skipping" % (byte_index))
                    byte_index += 1
                    continue

                stringBytes = file_bytes[stringStartInt : stringEndInt]
                stringText = ""
                # build the string one character at a time so that command strings can be escaped
                string_byte_index = 0
                readingSingleByte = False
                # single byte characters should be surrounded by [], while double-byte shift-jis should appear normally
                while string_byte_index < len(stringBytes):
                    if stringBytes[string_byte_index] < 0x80:
                        if readingSingleByte == False:
                            readingSingleByte = True
                            stringText += "["
                        stringText += chr(stringBytes[string_byte_index])
                        string_byte_index += 1

                    else: 
                        if readingSingleByte:
                            readingSingleByte = False
                            stringText += "]"

                        # hotfix to solve an issue with a Rafie dialog crash at 4486 @ file 001e0
                        # here, a single-byte shift-jis comma 82 F1 [A4] 88 EA is used instead of a double-byte comma
                        applyRafieFix = False
                        if string_byte_index > 2 and string_byte_index + 2 < len(stringBytes):
                            if stringBytes[string_byte_index] == 0xA4:
                                if stringBytes[string_byte_index-2] == 0x82:
                                    if stringBytes[string_byte_index-1] == 0xF1:
                                        if stringBytes[string_byte_index+1] == 0x88:
                                            if stringBytes[string_byte_index+2] == 0xEA:
                                                applyRafieFix = True

                        if applyRafieFix:
                            print("Rafie Single-Byte Comma Fix applied for string at position %d" % byte_index)
                            stringText += ","
                            string_byte_index += 1
                        else:
                            stringText += stringBytes[string_byte_index : string_byte_index + 2].decode(encoding='shift-jis')
                            string_byte_index += 2
                # close any open escapes
                if readingSingleByte:
                    readingSingleByte = False
                    stringText += "]"

                byteName = "95"
                if b == 0xA1:
                    byteName = "A1"

                shouldSkip = True
                if ignoreEmptyLines:
                    inBracket = False
                    for char in stringText:
                        if char == "[":
                            inBracket = True
                        elif char == "]":
                            inBracket = False
                        elif inBracket == False:
                            shouldSkip = False # some real text exists! don't skip

                if shouldSkip == True or (b == 0xA1 and stringBytes[0] == 0x00 and len(stringBytes) == 1):
                    # Skip zero-length string
                    pass
                #print(stringStartInt)
                else:
                    csv_writer.writerow([stringStartInt, stringEndInt, byteName, stringText, ''])

                byte_index = stringEndInt # jump over the entire string now that you've read it
            else: # anything else
                opcodeInfo = script_tools.informationForOpcodeAtLocation(file_bytes, byte_index, config, inputFilenameOnly, None, {})
                if opcodeInfo.valid:
                    byte_index += opcodeInfo.totalLength()
                else:
                    if b >= 0x0: 
                        print("WARNING: unexpected opcode %02x at position %d. Skipping." % (b, byte_index))

                        prevOpcodes = prevOpcodes[-15:]
                        opcodeHex = []
                        for opcode in prevOpcodes:
                            opcodeHex.append("%02x" % opcode)
                        print("         Previous opcodes: %s" % opcodeHex)
                    byte_index += 1

        output_file.close()
    print("Script unpacking of %s complete in %s seconds.\n" % (inputFilenameOnly, time.time() - start_time))

if __name__ == '__main__':
    main()
